interface PropertiesOptions {
    /** Default properties */
    defaults?: Properties | Map<string, string> | null;
    /** Variable regular expression */
    variableRegex?: RegExp | null;
    /** Line feed for stringify (default: \n) */
    lineFeed?: string;
    /** Specify separators used to split key-value pairs, default is [':', '='] */
    separators?: string[];
    /** Specify comment prefixes, default is ['#', '!'] */
    commentPrefixes?: string[];
}
export interface PropertiesConfig extends PropertiesOptions {
    props?: Map<string, string>;
    raw?: Map<string, string>;
}
export declare class Properties extends Map<string, string> {
    /**
     * @deprecated Use `setProperty()` instead
     */
    set: (key: string, value: string) => this;
    /**
     * @deprecated Use `getProperty()` instead
     */
    get: (key: string) => string | undefined;
    private readonly raw;
    private readonly options;
    constructor({ defaults, variableRegex, lineFeed, separators, commentPrefixes, props, raw }?: PropertiesConfig);
    get [Symbol.toStringTag](): string;
    setProperty(key: string, value: string): this;
    getProperty<T = string>(key: string, defaultValue?: T): string | undefined | T;
    getNumber(key: string, defaultValue?: number): number;
    getBoolean(key: string, defaultValue?: boolean): boolean;
    remove(key: string, specifiedValue?: string): boolean;
    isEmpty(): boolean;
    replace(key: string, oldValue: string, newValue?: string): boolean;
    clone(): Properties;
    parse(content: string, clear?: boolean): void;
    stringify(comments?: string, escapeUnicode?: boolean): string;
    toString(): string;
}
export {};
