(function(global, factory) {
  typeof exports === "object" && typeof module !== "undefined" ? factory(exports) : typeof define === "function" && define.amd ? define(["exports"], factory) : (global = typeof globalThis !== "undefined" ? globalThis : global || self, factory(global.PropertiesLike = {}));
})(this, function(exports2) {
  "use strict";
  const WHITE_SPACE = {
    " ": true,
    // 空格符
    "	": true,
    // 水平制表符
    "\f": true
    // 换页符
  };
  const LINE_FEED = {
    "\r": true,
    // 回车符
    "\n": true
    // 换行符
  };
  function noop() {
  }
  function parse(input, { separators = ["=", ":"], commentPrefixes = ["#", "!"], onData = noop, onComment = noop }) {
    const limit = input.length;
    let cursor = 0;
    let lineno = 1;
    let currentKey = "";
    let currentValue = "";
    let hasSep = false;
    let precedingBackslash = false;
    let skipWhiteSpace = true;
    let appendedLineBegin = false;
    let commentLine = "";
    let c;
    const collectKVPairs = (c2) => {
      if (!hasSep) {
        currentKey += c2;
      } else {
        currentValue += c2;
      }
    };
    while (cursor < limit) {
      c = input[cursor++];
      if (skipWhiteSpace) {
        if (WHITE_SPACE.hasOwnProperty(c)) {
          continue;
        }
        if (!appendedLineBegin && LINE_FEED.hasOwnProperty(c)) {
          lineno++;
          continue;
        }
        skipWhiteSpace = false;
        appendedLineBegin = false;
      }
      if (commentPrefixes.includes(c) && !precedingBackslash) {
        commentLine += c;
        do {
          c = input[cursor++];
          if (c <= "\r" && LINE_FEED.hasOwnProperty(c)) {
            skipWhiteSpace = true;
            break;
          }
          commentLine += c;
        } while (cursor < limit);
        onComment(commentLine, lineno);
        commentLine = "";
        lineno++;
        continue;
      }
      if (!LINE_FEED.hasOwnProperty(c)) {
        if (separators.includes(c)) {
          if (precedingBackslash) {
            collectKVPairs(c);
          } else if (!hasSep) {
            hasSep = true;
          } else {
            throw new Error(`Multiple separators '${c}' at [row: ${lineno}].`);
          }
          precedingBackslash = false;
        } else if (WHITE_SPACE.hasOwnProperty(c)) {
          if (precedingBackslash) {
            collectKVPairs(c);
          }
          precedingBackslash = false;
        } else if (c === "\\") {
          if (precedingBackslash) {
            collectKVPairs(c);
          }
          precedingBackslash = !precedingBackslash;
        } else if (c === "u") {
          if (precedingBackslash) {
            let value = 0;
            for (let i = 0; i < 4; i++) {
              c = input[cursor++];
              switch (c) {
                case "0":
                case "1":
                case "2":
                case "3":
                case "4":
                case "5":
                case "6":
                case "7":
                case "8":
                case "9":
                  value = (value << 4) + c.charCodeAt(0) - 48;
                  break;
                case "a":
                case "b":
                case "c":
                case "d":
                case "e":
                case "f":
                  value = (value << 4) + 10 + c.charCodeAt(0) - 97;
                  break;
                case "A":
                case "B":
                case "C":
                case "D":
                case "E":
                case "F":
                  value = (value << 4) + 10 + c.charCodeAt(0) - 65;
                  break;
                default:
                  throw new TypeError(`Malformed \\uxxxx encoding at [row: ${lineno}].`);
              }
            }
            c = String.fromCharCode(value);
          }
          precedingBackslash = false;
          collectKVPairs(c);
        } else {
          precedingBackslash = false;
          collectKVPairs(c);
        }
        if (cursor >= limit) {
          onData(currentKey, currentValue, lineno);
        }
      } else {
        if (precedingBackslash) {
          appendedLineBegin = true;
          precedingBackslash = false;
          if (c === "\r" && input[cursor] === "\n") {
            cursor++;
          }
        } else {
          hasSep = false;
          onData(currentKey, currentValue, lineno);
          currentKey = "";
          currentValue = "";
        }
        skipWhiteSpace = true;
        lineno++;
      }
    }
  }
  const ESCAPE_MAP = {
    t: "	",
    r: "\r",
    n: "\n",
    f: "\f",
    "\\": "\\",
    " ": "\\ "
  };
  function saveConvert(theString, separators, escapeUnicode) {
    let out = "";
    const len = theString.length;
    for (let x = 0; x < len; x++) {
      const aChar = theString[x];
      let aCharCode;
      const c = ESCAPE_MAP[aChar];
      if (c !== void 0) {
        out += "\\";
        out += c;
      } else if (separators.includes(aChar)) {
        out += "\\";
        out += aChar;
      } else if (escapeUnicode && ((aCharCode = aChar.charCodeAt(0)) < 32 || aCharCode > 126)) {
        out += `\\u${aCharCode.toString(16).padStart(4, "0")}`;
      } else {
        out += aChar;
      }
    }
    return out;
  }
  const RAW_TYPE_COMMENT = "COMMENT";
  const RAW_TYPE_PROPERTY = "PROPERTY";
  const RAW_TYPE_NEW_PROPERTY = "NEW_PROPERTY";
  const falseLike = {
    false: true,
    null: true,
    undefined: true,
    0: true,
    NaN: true,
    "": true
  };
  class Properties extends Map {
    constructor({ defaults, variableRegex, lineFeed, separators, commentPrefixes, props, raw } = {}) {
      super(props);
      this.options = {
        defaults: defaults || null,
        variableRegex: variableRegex || /\$\{\s*([^}]+)\s*\}/g,
        lineFeed: lineFeed || "\n",
        separators: separators || ["=", ":"],
        commentPrefixes: commentPrefixes || ["#", "!"]
      };
      const _raw = /* @__PURE__ */ new Map();
      if (raw) {
        for (const [key, value] of raw) {
          _raw.set(key, value);
        }
      }
      this.raw = _raw;
    }
    get [Symbol.toStringTag]() {
      return this.toString();
    }
    setProperty(key, value) {
      const { raw } = this;
      this.set(key, value);
      raw.set(key, raw.has(key) ? RAW_TYPE_PROPERTY : RAW_TYPE_NEW_PROPERTY);
      return this;
    }
    getProperty(key, defaultValue) {
      const { options: { defaults, variableRegex } } = this;
      let value = this.get(key);
      if (value === void 0) {
        if (defaults === null) {
          return defaultValue;
        }
        value = defaults.get(key);
        if (value === void 0) {
          return defaultValue;
        }
      }
      return variableRegex ? value.replace(variableRegex, (match, variable) => {
        const val = this.get(variable);
        return val !== void 0 ? val : "";
      }) : value;
    }
    getNumber(key, defaultValue) {
      const value = parseFloat(this.getProperty(key));
      return Number.isNaN(value) ? defaultValue === void 0 ? NaN : defaultValue : value;
    }
    getBoolean(key, defaultValue) {
      const value = this.getProperty(key);
      return value === void 0 ? defaultValue === void 0 ? false : defaultValue : !falseLike[value];
    }
    remove(key, specifiedValue) {
      let result = false;
      const { raw } = this;
      if (specifiedValue === void 0) {
        result = this.delete(key);
      } else {
        for (const [k, v] of this) {
          if (k === key && v === specifiedValue) {
            result = this.delete(k);
            break;
          }
        }
      }
      if (result && raw.get(key) === RAW_TYPE_NEW_PROPERTY) {
        raw.delete(key);
      }
      return result;
    }
    isEmpty() {
      return this.size === 0;
    }
    replace(key, oldValue, newValue) {
      if (newValue === void 0) {
        if (this.has(key)) {
          this.set(key, oldValue);
          return true;
        }
      } else if (this.get(key) === oldValue) {
        this.set(key, newValue);
        return true;
      }
      return false;
    }
    clone() {
      return new Properties(Object.assign({
        props: this,
        raw: this.raw
      }, this.options));
    }
    parse(content, clear) {
      const { raw } = this;
      if (clear) {
        this.clear();
        raw.clear();
      }
      parse(content, {
        onData: (key, value) => {
          raw.set(key, RAW_TYPE_PROPERTY);
          this.set(key, value);
        },
        onComment(comment) {
          raw.set(comment, RAW_TYPE_COMMENT);
        }
      });
    }
    stringify(comments, escapeUnicode) {
      if (typeof comments === "boolean") {
        escapeUnicode = comments;
        comments = void 0;
      }
      escapeUnicode = escapeUnicode !== false;
      let out = comments || "";
      const { raw, options: { lineFeed, separators } } = this;
      if (out.length > 0) {
        out += lineFeed;
      }
      for (const [key, rawType] of raw) {
        out += rawType === RAW_TYPE_COMMENT ? `${key}${lineFeed}` : `${saveConvert(key, separators, escapeUnicode)}=${saveConvert(this.get(key) || "", separators, escapeUnicode)}${lineFeed}`;
      }
      return out;
    }
    toString() {
      return "[object Properties]";
    }
  }
  exports2.Properties = Properties;
  exports2.parse = parse;
  Object.defineProperty(exports2, Symbol.toStringTag, { value: "Module" });
});
