const WHITE_SPACE = {
  " ": true,
  // 空格符
  "	": true,
  // 水平制表符
  "\f": true
  // 换页符
};
const LINE_FEED = {
  "\r": true,
  // 回车符
  "\n": true
  // 换行符
};
function noop() {
}
function parse(input, { separators = ["=", ":"], commentPrefixes = ["#", "!"], onData = noop, onComment = noop }) {
  const limit = input.length;
  let cursor = 0;
  let lineno = 1;
  let currentKey = "";
  let currentValue = "";
  let hasSep = false;
  let precedingBackslash = false;
  let skipWhiteSpace = true;
  let appendedLineBegin = false;
  let commentLine = "";
  let c;
  const collectKVPairs = (c2) => {
    if (!hasSep) {
      currentKey += c2;
    } else {
      currentValue += c2;
    }
  };
  while (cursor < limit) {
    c = input[cursor++];
    if (skipWhiteSpace) {
      if (WHITE_SPACE.hasOwnProperty(c)) {
        continue;
      }
      if (!appendedLineBegin && LINE_FEED.hasOwnProperty(c)) {
        lineno++;
        continue;
      }
      skipWhiteSpace = false;
      appendedLineBegin = false;
    }
    if (commentPrefixes.includes(c) && !precedingBackslash) {
      commentLine += c;
      do {
        c = input[cursor++];
        if (c <= "\r" && LINE_FEED.hasOwnProperty(c)) {
          skipWhiteSpace = true;
          break;
        }
        commentLine += c;
      } while (cursor < limit);
      onComment(commentLine, lineno);
      commentLine = "";
      lineno++;
      continue;
    }
    if (!LINE_FEED.hasOwnProperty(c)) {
      if (separators.includes(c)) {
        if (precedingBackslash) {
          collectKVPairs(c);
        } else if (!hasSep) {
          hasSep = true;
        } else {
          throw new Error(`Multiple separators '${c}' at [row: ${lineno}].`);
        }
        precedingBackslash = false;
      } else if (WHITE_SPACE.hasOwnProperty(c)) {
        if (precedingBackslash) {
          collectKVPairs(c);
        }
        precedingBackslash = false;
      } else if (c === "\\") {
        if (precedingBackslash) {
          collectKVPairs(c);
        }
        precedingBackslash = !precedingBackslash;
      } else if (c === "u") {
        if (precedingBackslash) {
          let value = 0;
          for (let i = 0; i < 4; i++) {
            c = input[cursor++];
            switch (c) {
              case "0":
              case "1":
              case "2":
              case "3":
              case "4":
              case "5":
              case "6":
              case "7":
              case "8":
              case "9":
                value = (value << 4) + c.charCodeAt(0) - 48;
                break;
              case "a":
              case "b":
              case "c":
              case "d":
              case "e":
              case "f":
                value = (value << 4) + 10 + c.charCodeAt(0) - 97;
                break;
              case "A":
              case "B":
              case "C":
              case "D":
              case "E":
              case "F":
                value = (value << 4) + 10 + c.charCodeAt(0) - 65;
                break;
              default:
                throw new TypeError(`Malformed \\uxxxx encoding at [row: ${lineno}].`);
            }
          }
          c = String.fromCharCode(value);
        }
        precedingBackslash = false;
        collectKVPairs(c);
      } else {
        precedingBackslash = false;
        collectKVPairs(c);
      }
      if (cursor >= limit) {
        onData(currentKey, currentValue, lineno);
      }
    } else {
      if (precedingBackslash) {
        appendedLineBegin = true;
        precedingBackslash = false;
        if (c === "\r" && input[cursor] === "\n") {
          cursor++;
        }
      } else {
        hasSep = false;
        onData(currentKey, currentValue, lineno);
        currentKey = "";
        currentValue = "";
      }
      skipWhiteSpace = true;
      lineno++;
    }
  }
}
export {
  parse
};
